package org.sun.jsat.trees;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import securityanalyzer.resourcemodel.FilledActionDefinition;
import securityanalyzer.resourcemodel.FilledRuleDefinition;
import securityanalyzer.resourcemodel.ParameterDefinition;
import securityanalyzer.resourcemodel.ResourceDefinition;
import securityanalyzer.useframework.CodeGrouping;
import securityanalyzer.useframework.RulesActionsContainer;

public class CodeGroupingRootTreeNode extends RefreshableTreeNode<CodeGrouping> {

    private Map<ResourceDefinition, RefreshableTreeNode<ResourceDefinition>> resourceGroupingMap = 
            new HashMap<ResourceDefinition, RefreshableTreeNode<ResourceDefinition>>();
    
    private Map<ResourceGroupDef, RefreshableTreeNode> parameterGroupingMap = 
            new HashMap<ResourceGroupDef, RefreshableTreeNode>();
    
    private int policyType;
    private String policyName;
    private int displayType;
    
    public static final int POLICY_TYPE_DEFAULT = 1;
    public static final int POLICY_TYPE_ANALYSIS = 2;
    public static final int POLICY_TYPE_EXTERNAL = 3;
    
    public static final int DISPLAY_ACTIONS = 1;
    public static final int DISPLAY_RULES = 2;
    
    public CodeGroupingRootTreeNode(CodeGrouping cg, int displayType) {
        super(cg, "Code Group");
        this.baseObject = cg;
        this.policyType = POLICY_TYPE_DEFAULT;
        this.policyName = null;
        this.displayType = displayType;
    }
    
    public CodeGroupingRootTreeNode(CodeGrouping cg, int displayType, int policyViewType) {
        super(cg, "Code Group");
        this.baseObject = cg;
        this.policyType = policyViewType;
        this.policyName = null;
        this.displayType = displayType;        
    }
    
    public CodeGroupingRootTreeNode(CodeGrouping cg, int displayType, int policyViewType, String policyName) {
        super(cg, "Code Group");
        this.baseObject = cg;
        this.policyType = policyViewType;
        this.policyName = policyName;
        this.displayType = displayType;
    }

    @Override
    public void generateMissingChildren() {
        synchronized (updateMutex){
            generationActions();
            if (baseObject == null) return;
            RulesActionsContainer rac = getProperRAC();
            switch (displayType) {
                case DISPLAY_ACTIONS:
                    for (FilledActionDefinition fad : rac.getActions()) {
                        RefreshableTreeNode acParent = 
                                getOrAddProperParent(fad);
                        if (childrenMap.get(fad) == null) {
                            FilledActionDefTreeNode fadtn = new FilledActionDefTreeNode(fad);
                            childrenMap.put(fad, fadtn);
                            acParent.add(fadtn);
                        }
                    }
                    break;
                case DISPLAY_RULES:
                    for (FilledRuleDefinition frd : rac.getRules()) {
                        if (childrenMap.get(frd) == null) {
                            FilledRuleDefTreeNode frdtn = new FilledRuleDefTreeNode(frd);
                            childrenMap.put(frd, frdtn);
                            this.add(frdtn);
                        }
                    }
                    break;
            }
        }
    }

    @Override
    public void removeExtraChildren() {
        synchronized (updateMutex) {
            generationActions();
            if (baseObject == null) return;
            RulesActionsContainer rac = getProperRAC();
            switch (displayType) {
                case DISPLAY_ACTIONS:
                    List<FilledActionDefinition> fads = new ArrayList<FilledActionDefinition>();
                    for (Entry<Object, RefreshableTreeNode> e : childrenMap.entrySet()) {
                        if (e.getKey() instanceof FilledActionDefinition) {
                            FilledActionDefinition a = (FilledActionDefinition) e.getKey();
                            if (!rac.getActions().contains(a)) {
                                fads.add(a);
                                e.getValue().removeFromParent();
                            }
                        }
                    }
                    for(FilledActionDefinition fad: fads) {
                        childrenMap.remove(fad);
                    }
                    cleanupGroups();
                    break;
                case DISPLAY_RULES:
                    List<FilledRuleDefinition> frds = new ArrayList<FilledRuleDefinition>();
                    for (Entry<Object, RefreshableTreeNode> e : childrenMap.entrySet()) {
                        if (e.getKey() instanceof FilledRuleDefinition) {
                            FilledRuleDefinition r = (FilledRuleDefinition) e.getKey();
                            if (!rac.getRules().contains(r)) {
                                frds.add(r);
                                this.remove(e.getValue());
                            }
                        }
                    }
                    for(FilledRuleDefinition frd: frds) {
                        childrenMap.remove(frd);
                    }
                    break;
            }
        }
    }

    @Override
    public void refresh() {
        super.refresh();
        if (displayType == DISPLAY_RULES){
            for (Object o : this.children){
                if (o instanceof FilledRuleDefTreeNode){
                    FilledRuleDefTreeNode frdtn = (FilledRuleDefTreeNode)o;
                    RulesActionsContainer rac = getProperRAC();
                    if (rac.getUnmatchedRules().contains(frdtn.baseObject)){
                        frdtn.setHighlight(true);
                    } else {
                        frdtn.setHighlight(false);
                    }
                }
            }
        }
    }
    
    

    public void setCg(CodeGrouping cg) {
        this.baseObject = cg;
    }
    
    public CodeGrouping getCg(){
        return baseObject;
    }

    private void cleanupGroups() {
        List<ResourceGroupDef> remLs = new ArrayList<ResourceGroupDef>();
        for (Entry<ResourceGroupDef, RefreshableTreeNode> e : parameterGroupingMap.entrySet()){
            
            if (e.getValue().getChildCount() <= 0){
                e.getValue().removeFromParent();
                remLs.add(e.getKey());
            }
            
        }
        for (ResourceGroupDef rgd : remLs){
                parameterGroupingMap.remove(rgd);
            }
        
        List<ResourceDefinition> remaLs = new ArrayList<ResourceDefinition>();
        for (Entry<ResourceDefinition, RefreshableTreeNode<ResourceDefinition>> e : resourceGroupingMap.entrySet()){
            
            if (e.getValue().getChildCount() <= 0){
                e.getValue().removeFromParent();
                remaLs.add(e.getKey());
            }
            
        }
        for (ResourceDefinition rgd : remaLs){
                resourceGroupingMap.remove(rgd);
            }
    }

    private RefreshableTreeNode<ResourceDefinition> getOrAddResourceGrouping(ResourceDefinition resource) {
        RefreshableTreeNode<ResourceDefinition> ret = resourceGroupingMap.get(resource);
        if (ret == null){
            ret = new RefreshableTreeNode<ResourceDefinition>(resource, resource.getName());
            resourceGroupingMap.put(resource, ret);
            this.add(ret);
        }
        return ret;
    }

    private RefreshableTreeNode getOrAddProperParent(FilledActionDefinition fad) {
        ResourceDefinition rd = fad.getBaseObject().getResource();
        RefreshableTreeNode<ResourceDefinition> resourceParent = getOrAddResourceGrouping(rd);
        ParameterDefinition gpd = fad.getBaseObject().getGroupBy();
        if (gpd == null) return resourceParent;
        
        String grouping = fad.getMapping(gpd);
        if (grouping == null) return resourceParent;
        
        ResourceGroupDef rgd = new ResourceGroupDef(rd, grouping);
        RefreshableTreeNode groupParent = parameterGroupingMap.get(rgd);
        if (groupParent == null){
            groupParent = new RefreshableTreeNode(grouping, grouping);
            resourceParent.add(groupParent);
            parameterGroupingMap.put(rgd, groupParent);
        }
        return groupParent;
    }

    private RulesActionsContainer getProperRAC(){
        RulesActionsContainer rac = null;
            switch (policyType){
                case 1:
                    rac = baseObject.getCurrentPolicy();
                    break;
                case 2:
                    rac = baseObject.getAnalysisPolicyByName(policyName);
                    break;
                case 3:
                    rac = baseObject.getExternalPolicyByName(policyName);
                    break;
                default:
                    //Invalid type...
                    return null;                            
            }
            return rac;
    }
    
    private class ResourceGroupDef {
        public ResourceDefinition resource = null;
        public String group = null;
        
        public ResourceGroupDef(ResourceDefinition rd, String groupName){
            this.resource = rd;
            this.group = groupName;
        }

        @Override
        public int hashCode() {
            return resource.hashCode() + group.hashCode();
        }

        @Override
        public boolean equals(Object obj) {
            if (!(obj instanceof ResourceGroupDef)) return false;
            return this.hashCode() == obj.hashCode();
        }
        
        
    }
}
