package securityanalyzer;

import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.StringTokenizer;

import securityanalyzer.analysismodel.AnalysisModel;
import securityanalyzer.analysismodel.CodeBlock;
import securityanalyzer.analysismodel.PermissionDescription;
import securityanalyzer.analysismodel.tools.soot.SootAnalysisTool;

/**
 * 
 * Main class for launching the SecurityAnalyzer - a tool to help developers
 * determine the permission requirements for their programs
 * 
 */
public class SecurityAnalyzer {

	static String[] availableToolsStrings = { "1" };
	static List<String> availableTools;

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		availableTools = Arrays.asList(availableToolsStrings);
		commandLoop();
	}

	public static void printHelp() {
		System.out.println();
		System.out.println("SecurityAnalyzer help...");
		System.out.println("help \n\t\tShow this list");
		System.out
				.println("run [k] [dir] [mainclass] \n\t\tRuns tool #k on the specified directory and main class.");
		System.out.println("tools \n\t\tLists available tools");
	}

	public static void printTools() {
		System.out.println("1: SootAnalysisTool");
	}

	public static void commandLoop() {
		BufferedReader buf = new BufferedReader(
				new InputStreamReader(System.in));
		String input = "";

		while (true) {
			System.out.print("SecurityAnalyzer :> ");

			try {
				input = buf.readLine();
			} catch (IOException e) {
				System.out
						.println("Definitely didn't expect an IOException.  Exiting...");
				System.exit(0);
			}

			processCommand(input.substring(input.lastIndexOf(">") + 1));
		}
	}

	public static void processCommand(String command) {

		StringTokenizer token = new StringTokenizer(command, " ");

		String commandType = "";

		try {
			commandType = token.nextToken();
		} catch (Exception e) {
			return;
		}

		if (commandType.equals("help")) {
			printHelp();
		} else if (commandType.equals("exit") || commandType.equals("quit")) {
			System.out.println("Exiting SecurityAnalyzer.");
			System.exit(0);
		} else if (commandType.equals("tools")) {
			printTools();
		} else if (commandType.equals("run")) {
			String[] runArgs = new String[3];

			int i=0;
			while (i < 3) {
				try {
					runArgs[i++] = token.nextToken();
				} catch (Exception e) {
					System.out
							.println("Bad arguments to run.  Type 'tools' to see a list.");
					return;
				}
			}



			runTool(runArgs);

		}

	}

	public static void runTool(String[] runArgs) {

		int tool = 0;

		if (!availableTools.contains(runArgs[0])) {
			System.out
					.println(runArgs[0]
							+ " is not a tool choice.  Type 'tools' to see a list.");
			return;

		}

		try {
			tool = Integer.parseInt(runArgs[0]);
		} catch (Exception e) {
			System.out
					.println(runArgs[0]
							+ " is not a tool choice.  Type 'tools' to see a list.");
			return;
		}
		
		if (tool == 1) {

			SootAnalysisTool sat = new SootAnalysisTool(
					runArgs[1], runArgs[2]);

			sat.runAnalysis();

			Collection<CodeBlock> results = AnalysisModel.getAllCodeBlocks();

			PrintStream fout = System.out;

			try {
				FileOutputStream fstream = new FileOutputStream("results.txt");
				fout = new PrintStream(fstream);
			} catch (Exception e) {
				System.out
						.println("Couldn't create file 'results.txt'... Printing output to console.");
			}

			for (CodeBlock cb : results) {
				if (cb.getType().equalsIgnoreCase("Method") && cb.isAppCode()) {
					printMethodResults(cb, fout);
				}
			}
		}

	}

	private static void printMethodResults(CodeBlock cb, PrintStream fout) {
		fout.println("-----------------------------------------------");
		fout.println("Permissions required for method with signature:");
		fout.println(cb.getFullyQualifiedName());

		Collection<PermissionDescription> pds = cb
				.getAllPermissionRequirements();

		if (pds.isEmpty()) {
			fout.println("No permissions found for "
					+ cb.getFullyQualifiedName());
		}

		for (PermissionDescription pd : pds) {
			fout.println("\t" + pd.getPermissionClass() + "("
					+ getPermissionArgString(pd) + ")");
		}

		fout.println("-----------------------------------------------");

	}

	private static String getPermissionArgString(PermissionDescription pd) {
		return Arrays.asList(pd.getPermissionArgs()).toString();
	}

}
