package securityanalyzer.analysismodel;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

/**
 * 
 * Defines a block of code. This can represent anything from a Method, to a
 * Class, to a Protection Domain. This also keeps track of the permission
 * requirements of the code being represented.
 * 
 */
public class CodeBlock {

	/** The fully qualified name of this bunch of code */
	private String fullyQualifiedName;
	/** The type of this container (i.e. Class, Package, etc) */
	private String type; //TODO is this useful?

	/** Permissions required for this bunch of code */
	private Set<PermissionDescription> permissionRequirements;
	/** Policies applied to this bunch of code */
	private Set<PolicyDescription> policiesApplied;

	/** The parent of this CodeBlock (or null if none) */
	private CodeBlock parent = null;
	/** The children of this CodeBlock */
	private Set<CodeBlock> children;

	/** True if the CodeBlock is in application code, false otherwise */
	private boolean isApp;

	/**
	 * Constructor
	 * @param type The type of this container (i.e. Class, Package, etc) 
	 * @param fullyQualifiedName The fully qualified name of this bunch of code 
	 */
	public CodeBlock(String type, String fullyQualifiedName) {
		this.permissionRequirements = new HashSet<PermissionDescription>();
		this.children = new HashSet<CodeBlock>();
		this.fullyQualifiedName = fullyQualifiedName;
		this.type = type;
		isApp = false;
	}

	/**
	 * 
	 */
	public void setAppCode(boolean setting) {
		this.isApp = setting;
	}

	public boolean isAppCode() {
		return isApp;
	}

	/**
	 * Use a PolicyApplicationStrategy as a Visitor to this CodeBlock
	 * @param pas
	 */
	public void applyStrategy(PolicyApplicationStrategy pas) {
		pas.process(this);
	}

	/**
	 * Get the FQN of the code represented by this object
	 * @return
	 */
	public String getFullyQualifiedName() {
		return fullyQualifiedName;
	}

	/**
	 * Get all permissions required by the code represented by this object
	 * @return
	 */
	public Set<PermissionDescription> getAllPermissionRequirements() {
		return permissionRequirements;
	}

	/**
	 * Get all policies applied to the code represented by this object
	 * @return
	 */
	public Set<PolicyDescription> getAllPoliciesApplied() {
		return policiesApplied;
	}

	/**
	 * Get the type of this CodeBlock
	 * @return
	 */
	public String getType() {
		return type;
	}

	/**
	 * Find a permission requirement of the specified Class and Name. If one
	 * does not exist, create one and return that.
	 * 
	 * @param permissionClass
	 *            The name of the permission class
	 * @param permissionName
	 *            The name of the permission
	 * @return The proper permission description
	 */
	public PermissionDescription getOrAddPermissionRequirement(
			String permissionClass, String[] args) {
		PermissionDescription p = getPermissionRequirement(permissionClass,
				args);
		if (p == null) {
			p = new PermissionDescription(permissionClass, args);
		}
		return p;
	}

	/**
	 * Add a set of permission requirements all at once
	 * 
	 */
	public void addPermissionRequirements(Set<PermissionDescription> pds) {
		permissionRequirements.addAll(pds);
	}

	/**
	 * Find a permission requirement of the specified Class and Name.
	 * 
	 * @param permissionClass
	 *            The name of the permission class
	 * @param permissionName
	 *            The name of the permission
	 * @return The proper permission description
	 */
	public PermissionDescription getPermissionRequirement(
			String permissionClass, String[] args) {
		for (PermissionDescription p : permissionRequirements) {
			if (p.getPermissionClass().equals(permissionClass)
					&& Arrays.equals(p.getPermissionArgs(), args)) {
				return p;
			}
		}
		return null;
	}

	/**
	 * Set the parent of this CodeBlock
	 * @param parent
	 */
	public void setParent(CodeBlock parent) {
		this.parent = parent;
	}

	/**
	 * Get the parent of this CodeBlock
	 * @return
	 */
	public CodeBlock getParent() {
		return parent;
	}

	/**
	 * Add a child to this CodeBlock
	 * @param child
	 */
	public void addChild(CodeBlock child) {
		this.children.add(child);
	}

	public boolean removeChild(CodeBlock child) {
		return this.children.remove(child);
	}

	/**
	 * Get all children of this CodeBlock
	 * @return
	 */
	public Set<CodeBlock> getChildren() {
		return children;
	}
}
