package securityanalyzer.analysismodel;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * 
 * Describes the requirement of a permission
 *
 */
public class PermissionRequirement {

	/** The permission being required */
	private PermissionDescription permission;

	/**
	 * All paths to sources of this permission requirement. Keyed by tool name.
	 * Value is a set of all the paths to this permission detected by the given
	 * tool (there can be many paths to the same permission, especially if the
	 * CodeBlock contains many methods)
	 */
	private Map<List<CodeBlock>, Set<String>> pathsToSource;
	// TODO - a path to a permission probably deserves its own class, with
	// conditions and such

	/**
	 * Constructor
	 * @param permission The permission being required
	 * @param path The path to get to the permission
	 * @param tool The tool that has detected the requirement
	 */
	public PermissionRequirement(PermissionDescription permission,
			List<CodeBlock> path, String tool) {
		this.permission = permission;
		this.pathsToSource = new HashMap<List<CodeBlock>, Set<String>>();
		this.addFinding(tool, path);
	}

	/**
	 * Get the permission being required
	 * @return
	 */
	public PermissionDescription getPermission() {
		return permission;
	}

	/**
	 * Get all paths by which this permission could be required
	 * @return
	 */
	public Collection<List<CodeBlock>> getAllPathsToSource() {
		Collection<List<CodeBlock>> ret = new HashSet<List<CodeBlock>>();
		for (List<CodeBlock> s : pathsToSource.keySet()) {
			ret.add(s);
		}
		return ret;
	}

	/**
	 * Get all paths by which this permission could be required found by the
	 * given tool
	 * 
	 * @param toolName
	 *            The name of the tool to filter by
	 * @return
	 */
	public Collection<List<CodeBlock>> getPathsToSource(String toolName) {
		Set<List<CodeBlock>> tempPaths = new HashSet<List<CodeBlock>>();
		
		for(List<CodeBlock> lobc: pathsToSource.keySet()) {
			if(pathsToSource.get(lobc).contains(toolName)) {
				tempPaths.add(lobc);
			}
		}
		
		return tempPaths;
	}

	/**
	 * Add a new path by which this permission could be required by
	 * 
	 * @param toolName
	 *            The name of the tool that found this path
	 * @param path
	 *            The path that was found
	 */
	public boolean addFinding(String toolName, List<CodeBlock> path) {
		Set<String> currTools = pathsToSource.get(path);
		
		if(currTools == null) {
			currTools = new HashSet<String>();
			pathsToSource.put(path, currTools);
		}
		
		return currTools.add(toolName);
	}
}
