package securityanalyzer.analysismodel.tools.soot;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import securityanalyzer.analysismodel.PermissionDescription;
import soot.Local;
import soot.PatchingChain;
import soot.RefType;
import soot.Scene;
import soot.SootField;
import soot.SootMethod;
import soot.Unit;
import soot.Value;
import soot.ValueBox;
import soot.jimple.InstanceInvokeExpr;
import soot.jimple.InvokeExpr;
import soot.jimple.StaticFieldRef;
import soot.jimple.Stmt;
import soot.jimple.StringConstant;
import soot.jimple.internal.JAssignStmt;
import soot.jimple.internal.JInvokeStmt;
import soot.jimple.toolkits.callgraph.CallGraph;
import soot.jimple.toolkits.callgraph.Edge;
import soot.tagkit.Tag;
import soot.util.Chain;

public class SootInitialPermissionCalculator {
	
	CallGraph cg;
	List<SootMethod> initialMethods;
	
	public SootInitialPermissionCalculator(CallGraph cg) {
		this.cg = cg;
		initialMethods = new ArrayList<SootMethod>();
	}
	
	public List<SootMethod> getInitialPermissions() {
		
		// Get the edges coming out of checkPermission()
		RefType rf = RefType.v("java.security.Permission");
		List<RefType> params = new ArrayList<RefType>();
		params.add(rf);
		SootMethod startM = Scene.v().getSootClass("java.lang.SecurityManager")
				.getMethod("checkPermission", params);

		Iterator<Edge> edges = 
			cg.edgesInto(startM);
		
		// Calculate Permissions
		setStartingPermissions(edges);
		
		return initialMethods;
		
	}
	
	public void setStartingPermissions(Iterator<Edge> edges) {

		while(edges.hasNext()) {
			Edge e = edges.next();

			Stmt s = e.srcStmt();

			ValueBox argBox = getArgBoxFromCall(s);

			if(argBox == null) {
				System.out.println("didn't have an invoke...");
				continue;
			}
			
			Value arg = argBox.getValue();

			Chain<Local> locals = e.src().getActiveBody().getLocals();

			PatchingChain<Unit> units = e.src().getActiveBody().getUnits();
			
			InstanceInvokeExpr ie = null;

			if (locals.contains(arg)) {
				// Need to start right before s
				ie = findInvokeExprOf(argBox, units, units.getPredOf(s));
			}

			String[] permArgs = getPermissionArgsFromExpr(ie);

			PermissionDescription pd = new PermissionDescription(arg.getType().toString(), permArgs);

			PermissionDescriptionTag pt = (PermissionDescriptionTag) e.src().getTag("PermissionDescriptionTag");

			if (pt == null) {
				pt = new PermissionDescriptionTag();
			}

			pt.addPermission(pd);

			e.src().addTag(pt);

			initialMethods.add(e.src());
		}   
	}
	
	public ValueBox getArgBoxFromCall(Stmt s) {

		if(s.containsInvokeExpr()) {
			InvokeExpr invoke = s.getInvokeExpr();
			return invoke.getArgBox(0);
		}

		return null;

	}

	public InstanceInvokeExpr findInvokeExprOf(ValueBox v, PatchingChain<Unit> units, Unit s) {

		
		// Start right before the relevant invocation of checkPermission
		Unit currUnit = s;
		
		boolean done=false;
		
		while(!done) {
			
			// Matters if its an invocation of (? extends Permission).<init>()
			if(currUnit instanceof JInvokeStmt) {
				JInvokeStmt jInv = (JInvokeStmt)currUnit;
				
				InvokeExpr inv = jInv.getInvokeExpr();
				
				// return if we are an instance of calling the constructor of the
				// permission (we already know that this is an invoke statement that
				// uses our arg that is of a permission type)
				if(inv instanceof InstanceInvokeExpr && inv.getMethodRef().name().equals("<init>")) {
					if(((InstanceInvokeExpr)inv).getBaseBox().getValue().equals(v.getValue())) {
						return (InstanceInvokeExpr) inv;
					}
				}
			}
			// The other case is if our argument is DEFINED here
			// Then we have to be a little more fancy, because we need
			// to see what it gets defined to and figure out what that
			// guy is doing.  Right now NOT worrying about branching
			// First case to consider is JAssignStmts, because they're
			// hopefully easier
			if(currUnit instanceof JAssignStmt) {
				JAssignStmt jAss = (JAssignStmt)currUnit;

				if(jAss.leftBox.getValue().equivTo(v.getValue())) {
					
					// If there is a static field reference, see if it can
					// be found in the static initializer
					if(jAss.rightBox.getValue() instanceof StaticFieldRef) {
						InstanceInvokeExpr inv = getStaticInvoke(jAss.rightBox);
						return inv;
					}
						
					// One idea is to just set v to be the rightBox...
					// then iterate again... and stop if we hit the beginning
					v = jAss.rightBox;
				}
			}
			
			if(units.getFirst().equals(currUnit)) {
				done = true;
			}
			else {
				currUnit = units.getPredOf(currUnit);
			}
		}
		
		return null;
		
		
	}
	
	public InstanceInvokeExpr getStaticInvoke(ValueBox vb) {
		
		SootField sf = ((StaticFieldRef)vb.getValue()).getField();
		
		// Must have a static initializer for this sf
		PatchingChain<Unit> units = sf.getDeclaringClass().getMethodByName("<clinit>").getActiveBody().getUnits();
		
		return findInvokeExprOf(vb, units, units.getLast());
	}
	
	@SuppressWarnings("unchecked")
	public String[] getPermissionArgsFromExpr(InstanceInvokeExpr ie) {

		if(ie == null) {
			return new String[0];
		}
		
		List args = ie.getArgs();

		List<String> fpArgs = new ArrayList<String>();

		for(Object permArg: args) {
			//PointsToAnalysis pta = Scene.v().getPointsToAnalysis();

			// If its a string constant, then just add that to the arg list
			// this is the most common (and possibly the only) case to
			// reasonably consider
			if(permArg instanceof StringConstant) {
				fpArgs.add(((StringConstant)permArg).value);
			}

			// A placeholder for args that weren't found
			else {
				fpArgs.add("???");
			}
			
		}   

		return fpArgs.toArray(new String[0]);

	}

}

class PermissionDescriptionTag implements Tag {
	HashSet<PermissionDescription> perms;
	
	public PermissionDescriptionTag() {
		perms = new HashSet<PermissionDescription>();
	}
	
	public PermissionDescriptionTag(HashSet<PermissionDescription> perms) {
		this.perms = perms;
	}
	
	public String getName() {
		return "PermissionDescriptionTag";
	}
	
	public byte[] getValue() {
		throw new UnsupportedOperationException();
	}
	
	public HashSet<PermissionDescription> getPermissions() {
		return perms;
	}
	
	public void addPermission(PermissionDescription pd) {
		perms.add(pd);
	}
	
	public boolean hasAllPermissionsOf(PermissionDescriptionTag pdt) {
		return perms.containsAll(pdt.getPermissions());
	}
	
	public void addPermissions(Collection<PermissionDescription> pds) {
		perms.addAll(pds);
	}
	
	public String toString() {
		String s = "";
		for(PermissionDescription pd: perms) {
			s += pd.toString() + "\n";
		}
		return s;
	}
}
