package securityanalyzer.analysismodel.tools.soot;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import soot.G;
import soot.PackManager;
import soot.Scene;
import soot.SceneTransformer;
import soot.SootClass;
import soot.SootMethod;
import soot.Transform;
import soot.jimple.toolkits.callgraph.CallGraph;
import soot.jimple.toolkits.callgraph.Edge;
import soot.options.Options;
import soot.util.Chain;
import soot.util.queue.QueueReader;

/**
 * Model of the program that is used to get initial state
 * of PermissionDescriptions and CallGraph to build the
 * AnalysisModel.
 * @author joe
 *
 */
public class SootStaticModel {
	
	final SootStaticModel instance = this;
	
	CallGraph callGraph;
	Chain<SootClass> classes;
	
	/**
	 * 
	 */
	public SootStaticModel() {
		// Set necessary options and get initial data structures from soot
		Options.v().set_whole_program(true);
		Options.v().setPhaseOption("cg.spark", "on");
	}
	
	/**
	 * Sets up Soot with options needed to run analysis
	 */
	public void init() {

		
		Scene.v().loadNecessaryClasses();
		
		PackManager.v().getPack("wjtp").add(new Transform("wjtp.myTrans", new SceneTransformer() {
			@SuppressWarnings("unchecked")
			@Override
			protected void internalTransform(String phaseName, Map options) {
				
				CallGraph cg = Scene.v().getCallGraph();
				Chain<SootClass> classes = Scene.v().getClasses();
				
				instance.setFields(cg, classes);
			}
		}));

		
		PackManager.v().getPack("cg").apply();
		PackManager.v().getPack("wjtp").apply();
	}
	
	/**
	 * Loads the specified class to the scene, and sets it as an application class
	 * 
	 * @param className
	 */
	public void addClassToScene(String className) {
		if(Scene.v().containsClass(className))
			return;
		SootClass sc = Scene.v().loadClassAndSupport(className);
		sc.setApplicationClass();
	}
	
	/**
	 * Adds the specified class if not already added
	 * and sets is as the main class
	 * @param mainClass
	 */
	public void setMainClass(String mainClass) {
		addClassToScene(mainClass);
		Options.v().set_main_class(mainClass);
	}
	
	private void setFields(CallGraph cg, Chain<SootClass> classes) {
		this.callGraph = cg;
		this.classes = classes;
	}
	
	public CallGraph getCallGraph() {
		return callGraph;
	}
	
	public Collection<Edge> getEdges() {
		QueueReader<Edge> edgeReader = callGraph.listener();
		Collection<Edge> edges = new ArrayList<Edge>();
		
		while(edgeReader.hasNext()) {
			edges.add(edgeReader.next());
			
		}
		
		return edges;
	}
	
	public Chain<SootClass> getClasses() {
		return classes;
	}
	
	public List<SootMethod> getInitialMethods() {
		SootInitialPermissionCalculator sipc = new SootInitialPermissionCalculator(callGraph);
		
		return sipc.getInitialPermissions();
		
	}
	
	/**
	 * Resets soot and gets rid of analysis data
	 */
	public void clearSootData() {
		
		Scene.v().releaseActiveHierarchy();
		Scene.v().releaseCallGraph();
		Scene.v().releaseReachableMethods();
		Scene.v().releasePointsToAnalysis();
		Scene.v().releaseFastHierarchy();
		Scene.v().releaseSideEffectAnalysis();
		
		G.reset();
	}

	public void setProcessDir(String processDir) {
		
		Options.v().set_process_dir(Arrays.asList(processDir));
		Options.v().set_prepend_classpath(true);
		Options.v().set_soot_classpath(processDir);
		
	}
	
	

}
