package securityanalyzer.resourcemodel;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.regex.Pattern;

import securityanalyzer.resourcemodel.ActionDefinition.AppliedRule;

public class VariableResolution {

	/**
	 * Resolves all parameters in a String to their proper values by replacing
	 * variables from the given parameter map. Variables are enclosed in '{' and
	 * '}'.
	 * 
	 * @param value
	 *            String to do replacement on
	 * @param mappings
	 *            Mappings from variable name to value
	 * @return
	 */
	public static String resolveVariables(String value,
			Map<String, String> mappings) {
		List<Integer> start = new LinkedList<Integer>();
		List<Integer> end = new LinkedList<Integer>();
	
		int i = 0;
		for (char c : value.toCharArray()) {
			if (c == '{') {
				start.add(i);
			} else if (c == '}') {
				end.add(i);
			}
			i++;
		}
	
		if (start.size() != end.size()) {
			// TODO : handle mismatched start and end pieces!
			return null;
		}
                
                for (int c = 0; c < start.size() - 1; c++){
                    //start.c+1 > end.c
                    if (start.get(c+1) < end.get(c)) return null;
                }
	
		StringBuffer ret = new StringBuffer();
		i = 0;
		int index = 0;
		for (Integer s : start) {
			while (i < s) {
				ret.append(value.charAt(i));
				i++;
			}
	
			Integer e = end.get(index);
			String var = mappings.get(value.substring(s + 1, e));
			if (var == null){
				//TODO handle missing vars
				var = "";
			}
			ret.append(var);
			i = e + 1;
			index++;
		}
		
		if (i < value.length()) {
			ret.append(value.substring(i));
		}
		
		
	
		return ret.toString();
	}

	
	public static Map<String, String> simplifyDefinitionValueMappings(
			Map<? extends Definition, String> defValMap) {
		Map<String, String> ret = new HashMap<String, String>();
		for (Entry<?extends Definition, String> e : defValMap.entrySet()) {
			ret.put(e.getKey().getName(), e.getValue());
		}
		return ret;
	}


	public static Set<FilledRuleDefinition> getFilledRulesFrom(
				FilledActionDefinition fad, String type) {
			Set<FilledRuleDefinition> ret = new HashSet<FilledRuleDefinition>();
	
			Map<ParameterDefinition, String> mappings = fad
					.getAllMappingsResolveDefaults();
			Map<String, String> actionParameterMap = new HashMap<String, String>();
	
			// Here we get a map of actionParameter names to their values
			for (Entry<ParameterDefinition, String> e : mappings.entrySet()) {
				actionParameterMap.put(e.getKey().getName(), e.getValue());
			}
	
			Set<AppliedRule> rules = fad.getBaseObject().getParameterApplications(
					type);
	
			//int i = 0;
	
			// Need to get the parameters' names and value strings from the applied
			// rules
			for (AppliedRule rule : rules) {
				FilledRuleDefinition frd = new FilledRuleDefinition(
						rule.ruleDefinition);
				Map<String, String> nameValueMap = rule.nameValueMap;
				// now resolve each entry in the nameValueMap
				for (ParameterDefinition pd : rule.ruleDefinition.getParameters()) {
					
					String appliedValue = resolveVariables(nameValueMap.get(pd.getName()), actionParameterMap);
					
					frd.setMapping(pd, appliedValue);
					
				}
				
				// and add it to the resulting list
				ret.add(frd);
			}
	
			return ret;
		}


	/**
	 * Does reverse parameter variable mapping to get the values of action
	 * parameters that would be used to specify the given rule parameter
	 * 
	 * @param template
	 *            The template for a rule parameter found in an Applied Rule
	 * @param ruleParameter
	 *            A defined parameter found in a FilledRuleDefinition
	 * @return A map of names of actionParameters to the values found in the
	 *         ruleParameter
	 */
	public static Map<String, String> getActionParametersFromRuleParameter(
			String template, String ruleParameter) {
	
		// Assumptions:
		// 1. Curly brackets are special and will only appear in templates when
		// indicating parameter variables
		// 2. Each variable name is only used once
	
		// Right now this doesn't handle badly formed relationships between
		// template
		// and ruleParameter very well
	
		// This was an interesting approach to this - invert the order of the
		// brackets and use regex split to find the variable names and the
		// tokens between them
		String inverseTemplate = VariableResolution.invertBrackets(template);
	
		// Get everything that isn't in between brackets
		Pattern p = Pattern.compile("\\{.*?\\}");
		String[] varNames = VariableResolution.removeEmptyStrings(p.split(inverseTemplate));
		String[] tokens = VariableResolution.removeEmptyStrings(p.split(template));
	
		Map<String, String> actionParams = new HashMap<String, String>();
	
		// If there are no variables to figure out, just return an empty map
		if (varNames.length == 0) {
			return actionParams;
		}
	
		// This is a little naive because the tokens could appear within the
		// values that we are trying to find, but it works for the cases tried
		// so far
	
		int i = 0, start = 0, end = 0, currVar = 0;
	
		// If there are no tokens found, then the whole thing is variables but
		// we assume there's only one, since there's no way to figure out where
		// the separation would be
		if (tokens.length == 0) {
			actionParams.put(varNames[0], ruleParameter);
			return actionParams;
		}
	
		// Check if the string started with the first token
		if (ruleParameter.startsWith(tokens[0])) {
			start = tokens[0].length();
			// This is to catch the case where the format is just [token]{var}
			end = start;
			i++;
		}
		
		// Now just keep looking at what is in between the tokens and extract
		// out what the variables are
		while (i < tokens.length) {
			end = ruleParameter.indexOf(tokens[i]);
			// If end is -1, we're looking at the wrong thing
			if(end == -1) {
				return new HashMap<String, String>();
			}
			actionParams.put(varNames[currVar++], ruleParameter.substring(
					start, end));
			start = end;
			start += tokens[i].length();
			i++;
		}
	
		// No more tokens, so if there's any variable left, it's the rest
		if (!ruleParameter.endsWith(tokens[i - 1])) {
			actionParams.put(varNames[currVar], ruleParameter.substring(end));
		}
	
		return actionParams;
	}


	/**
	 * Used for processing reverse mapping. Turns all '{' in s into '}' and vice
	 * versa
	 * 
	 * @param s
	 * @return
	 */
	public static String invertBrackets(String s) {
	
		String inverse = "";
	
		for (int j = 0; j < s.length(); j++) {
			if (s.charAt(j) == '}') {
				inverse += '{';
			} else if (s.charAt(j) == '{') {
				inverse += '}';
			} else {
				inverse += s.charAt(j);
			}
		}
	
		// If it starts with a close bracket, then need to chop it off
		if (inverse.startsWith("}")) {
			inverse = inverse.substring(1);
		}
		// otherwise, need the open bracket to match
		else {
			inverse = '{' + inverse;
		}
		// If it ends with an open bracket, then need to chop it off
		if (inverse.endsWith("{")) {
			inverse = inverse.substring(0, inverse.length() - 1);
		} else {
			inverse = inverse + '}';
		}
	
		return inverse;
	}


	/**
	 * Takes in an array of strings and returns an array of strings with all
	 * elements that were not the empty string in the same order
	 * 
	 * @param strings
	 * @return
	 */
	static String[] removeEmptyStrings(String[] strings) {
		List<String> newStrings = new ArrayList<String>();
	
		for (String s : strings) {
			if (!s.equals("")) {
				newStrings.add(s);
			}
		}
	
		return newStrings.toArray(new String[0]);
	}
}
